@@ -1,7 +1,7 @@
 '''
 dugong.py - Python HTTP Client Module
 
-Copyright (C) Nikolaus Rath <Nikolaus@rath.org>
+Copyright © 2014 Nikolaus Rath <Nikolaus.org>
 
 This module may be distributed under the terms of the Python Software Foundation
 License Version 2.
@@ -33,8 +33,13 @@ try:
 except ImportError:
     asyncio = None
 
+# Enums are only available in 3.4 and newer
+try:
+    from enum import Enum
+except ImportError:
+    Enum = object
 
-__version__ = '3.3'
+__version__ = '3.7.1'
 
 log = logging.getLogger(__name__)
 
@@ -50,15 +55,43 @@ MAX_LINE_SIZE = BUFFER_SIZE-1
 #: this value, `InvalidResponse` will be raised.
 MAX_HEADER_SIZE = BUFFER_SIZE-1
 
-CHUNKED_ENCODING = 'chunked_encoding'
-IDENTITY_ENCODING = 'identity_encoding'
+class Symbol:
+    '''
+    A symbol instance represents a specific state. Its value is
+    not relevant, as it should only ever be assigned to or compared
+    with other variables.
+    '''
+    __slots__ = [ 'name' ]
 
-#: Marker object for request body size when we're waiting
-#: for a 100-continue response from the server
-WAITING_FOR_100c = object()
+    def __init__(self, name):
+        self.name = name
+
+    def __str__(self):
+        return self.name
+
+    def __repr__(self):
+        return '<%s>' % (self.name,)
+
+class Encodings(Enum):
+    CHUNKED = 1
+    IDENTITY = 2
+
+#: Sentinel for `HTTPConnection._out_remaining` to indicate that
+#: we're waiting for a 100-continue response from the server
+WAITING_FOR_100c = Symbol('WAITING_FOR_100c')
+
+#: Sentinel for `HTTPConnection._in_remaining` to indicate that
+#: the response body cannot be read and that an exception
+#: (stored in the `HTTPConnection._encoding` attribute) should
+#: be raised.
+RESPONSE_BODY_ERROR = Symbol('RESPONSE_BODY_ERROR')
+
+#: Sentinel for `HTTPConnection._in_remaining` to indicate that
+#: we should read until EOF (i.e., no keep-alive)
+READ_UNTIL_EOF = Symbol('READ_UNTIL_EOF')
 
 #: Sequence of ``(hostname, port)`` tuples that are used by
-#: `is_temp_network_error` to distinguish between permanent and temporary name
+#: dugong to distinguish between permanent and temporary name
 #: resolution problems.
 DNS_TEST_HOSTNAMES=(('www.google.com', 80),
                     ('www.iana.org', 80),
@@ -144,7 +177,13 @@ class HTTPResponse:
         #: HTTP Response headers, a `email.message.Message` instance
         self.headers = headers
 
-        #: Length of the response body or `None`, if not known
+        #: Length of the response body or `None` if not known. This attribute
+        #: contains the actual length of the *transmitted* response. That means
+        #: that for responses where RFC 2616 mandates that no request body
+        #: be sent (e.g. in response to HEAD requests or for 1xx response
+        #: codes) this value is zero. In these cases, the length of the body that
+        #: *would* have been send can be extracted from the ``Content-Length``
+        #: response header.
         self.length = length
 
 
@@ -185,6 +224,36 @@ class _GeneralError(Exception):
     def __str__(self):
         return self.msg
 
+class HostnameNotResolvable(Exception):
+    '''Raised if a host name does not resolve to an ip address.
+
+    Dugong raises this exception if a resolution attempt results in a
+    `socket.gaierror` or `socket.herror` exception with errno
+    :const:`!socket.EAI_AGAIN` or :const:`!socket.EAI_NONAME`, but at least one of
+    the hostnames in `DNS_TEST_HOSTNAMES` can be resolved.
+    '''
+
+    def __init__(self, hostname):
+        self.name = hostname
+
+    def __str__(self):
+        return 'Host %s does not have any ip addresses' % self.name
+
+
+class DNSUnavailable(Exception):
+    '''Raised if the DNS server cannot be reached.
+
+    Dugong raises this exception if a resolution attempt results in a
+    `socket.gaierror` or `socket.herror` exception with errno
+    :const:`!socket.EAI_AGAIN` or :const:`!socket.EAI_NONAME`, and none of the
+    hostnames in `DNS_TEST_HOSTNAMES` can be resolved either.
+    '''
+
+    def __init__(self, hostname):
+        self.name = hostname
+
+    def __str__(self):
+        return 'Unable to resolve %s, DNS server unavailable.' % self.name
 
 class StateError(_GeneralError):
     '''
@@ -230,7 +299,18 @@ class UnsupportedResponse(_GeneralError):
 
 class ConnectionClosed(_GeneralError):
     '''
-    Raised if the server unexpectedly closed the connection.
+    Raised if the connection was unexpectedly closed.
+
+    This exception is raised also if the server declared that it will close the
+    connection (by sending a ``Connection: close`` header). Such responses can
+    still be read completely, but the next attempt to send a request or read a
+    response will raise the exception. To re-use the connection after the server
+    has closed the connection, call `HTTPConnection.reset` before further
+    requests are send.
+
+    This behavior is intentional, because the caller may have already issued
+    other requests (i.e., used pipelining). By raising an exception, the caller
+    is notified that any pending requests have been lost and need to be resend.
     '''
 
     msg = 'connection closed unexpectedly'
@@ -360,11 +440,13 @@ class HTTPConnection:
         #: request headers have been sent, but request body data is still
         #: pending, it is set to a ``(method, path, body_len)`` tuple. *body_len*
         #: is the number of bytes that that still need to send, or
-        #: WAITING_FOR_100c if we are waiting for a 100 response from the server.
+        #: `WAITING_FOR_100c` if we are waiting for a 100 response from the server.
         self._out_remaining = None
 
         #: Number of remaining bytes of the current response body (or current
-        #: chunk), or `None` if the response header has not yet been read.
+        #: chunk), `None` if there is no active response or `READ_UNTIL_EOF` if
+        #: we have to read until the connection is closed (i.e., we don't know
+        #: the content-length and keep-alive is not active).
         self._in_remaining = None
 
         #: Transfer encoding of the active response (if any).
@@ -401,22 +483,33 @@ class HTTPConnection:
 
         if self.proxy:
             log.debug('connecting to %s', self.proxy)
-            self._sock = socket.create_connection(self.proxy)
-            eval_coroutine(self._co_tunnel(), self.timeout)
+            self._sock = create_socket(self.proxy)
+            if self.ssl_context:
+                eval_coroutine(self._co_tunnel(), self.timeout)
         else:
             log.debug('connecting to %s', (self.hostname, self.port))
-            self._sock = socket.create_connection((self.hostname, self.port))
+            self._sock = create_socket((self.hostname, self.port))
 
         if self.ssl_context:
             log.debug('establishing ssl layer')
-            server_hostname = self.hostname if ssl.HAS_SNI else None
+            if (sys.version_info >= (3, 5) or
+                (sys.version_info >= (3, 4) and ssl.HAS_SNI)):
+                # Automatic hostname verification was added in 3.4, but only
+                # if SNI is available. In 3.5 the hostname can be checked even
+                # if there is no SNI support.
+                server_hostname = self.hostname
+            else:
+                server_hostname = None
             self._sock = self.ssl_context.wrap_socket(self._sock, server_hostname=server_hostname)
 
-            try:
-                ssl.match_hostname(self._sock.getpeercert(), self.hostname)
-            except:
-                self.close()
-                raise
+            if server_hostname is None:
+                # Manually check hostname for Python < 3.4, or if we have
+                # 3.4 without SNI.
+                try:
+                    ssl.match_hostname(self._sock.getpeercert(), self.hostname)
+                except:
+                    self.close()
+                    raise
 
         self._sock.setblocking(False)
         self._rbuf.clear()
@@ -552,7 +645,11 @@ class HTTPConnection:
         headers['Accept-Encoding'] = 'identity'
         if 'Connection' not in headers:
             headers['Connection'] = 'keep-alive'
-        request = [ '{} {} HTTP/1.1'.format(method, path).encode('latin1') ]
+        if self.proxy and not self.ssl_context:
+            gpath = "http://{}{}".format(headers['Host'], path)
+        else:
+            gpath = path
+        request = [ '{} {} HTTP/1.1'.format(method, gpath).encode('latin1') ]
         for key, val in headers.items():
             request.append('{}: {}'.format(key, val).encode('latin1'))
         request.append(b'')
@@ -577,10 +674,10 @@ class HTTPConnection:
         if not isinstance(buf, memoryview):
             buf = memoryview(buf)
 
-
-        fd = self._sock.fileno()
         while True:
             try:
+                if self._sock is None:
+                    raise ConnectionClosed('connection has been closed locally')
                 len_ = self._sock.send(buf)
                 # An SSL socket has the nasty habit of returning zero
                 # instead of raising an exception when in non-blocking
@@ -589,12 +686,10 @@ class HTTPConnection:
                     raise BlockingIOError()
             except (socket.timeout, ssl.SSLWantWriteError, BlockingIOError):
                 log.debug('yielding')
-                yield PollNeeded(fd, POLLOUT)
-                if self._sock is None:
-                    raise ConnectionClosed('connection has been closed locally')
+                yield PollNeeded(self._sock.fileno(), POLLOUT)
                 continue
             except (BrokenPipeError, ConnectionResetError):
-                raise ConnectionClosed('found closed when trying to write')
+                raise ConnectionClosed('connection was interrupted')
             except OSError as exc:
                 if exc.errno == errno.EINVAL:
                     # Blackhole routing, according to ip(7)
@@ -629,9 +724,6 @@ class HTTPConnection:
         if not self._out_remaining:
             raise StateError('No active request with pending body data')
 
-        if self._sock is None:
-            raise ConnectionClosed('connection has been closed locally')
-
         (method, path, remaining) = self._out_remaining
         if remaining is WAITING_FOR_100c:
             raise StateError("can't write when waiting for 100-continue")
@@ -682,9 +774,6 @@ class HTTPConnection:
 
         log.debug('start')
 
-        if self._sock is None:
-            raise ConnectionClosed('connection has been closed locally')
-
         if len(self._pending_requests) == 0:
             raise StateError('No pending requests')
 
@@ -737,57 +826,78 @@ class HTTPConnection:
         #
         # Prepare to read body
         #
-        body_length = None
+        body_length = self._setup_read(method, status, header)
 
-        tc = header['Transfer-Encoding']
-        if tc:
-            tc = tc.lower()
-        if tc and tc == 'chunked':
-            log.debug('Chunked encoding detected')
-            self._encoding = CHUNKED_ENCODING
-            self._in_remaining = 0
-        elif tc and tc != 'identity':
-            # Server must not sent anything other than identity or chunked, so
-            # we raise InvalidResponse rather than UnsupportedResponse. We defer
-            # raising the exception to read(), so that we can still return the
-            # headers and status (and don't fail if the response body is empty).
-            log.warning('Server uses invalid response encoding "%s"', tc)
-            self._encoding = InvalidResponse('Cannot handle %s encoding' % tc)
-        else:
-            log.debug('identity encoding detected')
-            self._encoding = IDENTITY_ENCODING
+        # Don't require calls to co_read() et al if there is
+        # nothing to be read.
+        if self._in_remaining is None:
+            self._pending_requests.popleft()
+
+        log.debug('done (in_remaining=%s)', self._in_remaining)
+        return HTTPResponse(method, path, status, reason, header, body_length)
+
+    def _setup_read(self, method, status, header):
+        '''Prepare for reading response body
+
+        Sets up `._encoding`, `_in_remaining` and returns Content-Length
+        (if available).
+
+        See RFC 2616, sec. 4.4 for specific rules.
+        '''
+
+        # On error, the exception is stored in _encoding and raised on
+        # the next call to co_read() et al - that way we can still
+        # return the http status and headers.
+
+        will_close = header.get('Connection', 'keep-alive').lower() == 'close'
+
+        body_length = header['Content-Length']
+        if body_length is not None:
+            try:
+                body_length = int(body_length)
+            except ValueError:
+                self._encoding = InvalidResponse('Invalid content-length: %s'
+                                                 % body_length)
+                self._in_remaining = RESPONSE_BODY_ERROR
+                return None
 
-        # does the body have a fixed length? (of zero)
         if (status == NO_CONTENT or status == NOT_MODIFIED or
             100 <= status < 200 or method == 'HEAD'):
             log.debug('no content by RFC')
-            body_length = 0
             self._in_remaining = None
-            self._pending_requests.popleft()
+            self._encoding = None
+            return 0
 
-        # Chunked doesn't require content-length
-        elif self._encoding is CHUNKED_ENCODING:
-            pass
-
-        # Otherwise we require a content-length. We defer raising the exception
-        # to read(), so that we can still return the headers and status.
-        elif ('Content-Length' not in header
-              and not isinstance(self._encoding, InvalidResponse)):
-            log.debug('no content length and no chunkend encoding, will raise on read')
-            self._encoding = UnsupportedResponse('No content-length and no chunked encoding')
+        tc = header.get('Transfer-Encoding', 'identity').lower()
+        if tc == 'chunked':
+            log.debug('Chunked encoding detected')
+            self._encoding = Encodings.CHUNKED
             self._in_remaining = 0
+            return None
 
-        else:
-            body_length = int(header['Content-Length'])
-            if body_length:
-                self._in_remaining = body_length
-            else:
-                self._in_remaining = None
-                self._pending_requests.popleft()
+        elif tc != 'identity':
+            log.warning('Server uses invalid response encoding "%s"', tc)
+            self._encoding = InvalidResponse('Cannot handle %s encoding' % tc)
+            self._in_remaining = RESPONSE_BODY_ERROR
+            return None
 
-        log.debug('done (in_remaining=%s)', self._in_remaining)
+        log.debug('identity encoding detected')
+        self._encoding = Encodings.IDENTITY
 
-        return HTTPResponse(method, path, status, reason, header, body_length)
+        if body_length is not None:
+            log.debug('Will read response body of %d bytes', body_length)
+            self._in_remaining = body_length or None
+            return body_length
+
+        if will_close:
+            log.debug('no content-length, will read until EOF')
+            self._in_remaining = READ_UNTIL_EOF
+            return None
+
+        log.debug('no content length and no chunked encoding, will raise on read')
+        self._encoding = UnsupportedResponse('No content-length and no chunked encoding')
+        self._in_remaining = RESPONSE_BODY_ERROR
+        return None
 
     def _co_read_status(self):
         '''Read response line'''
@@ -871,21 +981,16 @@ class HTTPConnection:
 
         log.debug('start (len=%d)', len_)
 
-        if self._sock is None:
-            raise ConnectionClosed('connection has been closed locally')
-
-        if len_ is None:
-            return (yield from self.co_readall())
-
-        if len_ == 0 or self._in_remaining is None:
-            return b''
-
-        if self._encoding is IDENTITY_ENCODING:
-            return (yield from self._co_read_id(len_))
-        elif self._encoding is CHUNKED_ENCODING:
-            return (yield from self._co_read_chunked(len_=len_))
-        elif isinstance(self._encoding, Exception):
+        if self._in_remaining is RESPONSE_BODY_ERROR:
             raise self._encoding
+        elif len_ is None:
+            return (yield from self.co_readall())
+        elif len_ == 0 or self._in_remaining is None:
+            return b''
+        elif self._encoding is Encodings.IDENTITY:
+            return (yield from self._co_read_id(len_))
+        elif self._encoding is Encodings.CHUNKED:
+            return (yield from self._co_read_chunked(len_=len_))
         else:
             raise RuntimeError('ooops, this should not be possible')
 
@@ -943,18 +1048,14 @@ class HTTPConnection:
 
         log.debug('start (buflen=%d)', len(buf))
 
-        if self._sock is None:
-            raise ConnectionClosed('connection has been closed locally')
-
-        if len(buf) == 0 or self._in_remaining is None:
-            return 0
-
-        if self._encoding is IDENTITY_ENCODING:
-            return (yield from self._co_readinto_id(buf))
-        elif self._encoding is CHUNKED_ENCODING:
-            return (yield from self._co_read_chunked(buf=buf))
-        elif isinstance(self._encoding, Exception):
+        if self._in_remaining is RESPONSE_BODY_ERROR:
             raise self._encoding
+        elif len(buf) == 0 or self._in_remaining is None:
+            return 0
+        elif self._encoding is Encodings.IDENTITY:
+            return (yield from self._co_readinto_id(buf))
+        elif self._encoding is Encodings.CHUNKED:
+            return (yield from self._co_read_chunked(buf=buf))
         else:
             raise RuntimeError('ooops, this should not be possible')
 
@@ -970,9 +1071,9 @@ class HTTPConnection:
             self._pending_requests.popleft()
             return b''
 
-        sock_fd = self._sock.fileno()
         rbuf = self._rbuf
-        len_ = min(len_, self._in_remaining)
+        if self._in_remaining is not READ_UNTIL_EOF:
+            len_ = min(len_, self._in_remaining)
         log.debug('updated len_=%d', len_)
 
         # If buffer is empty, reset so that we start filling from
@@ -987,17 +1088,24 @@ class HTTPConnection:
         # and buffer is not full
         while len(rbuf) < len_ and rbuf.e < len(rbuf.d):
             got_data = self._try_fill_buffer()
-            if not got_data and not rbuf:
-                log.debug('buffer empty and nothing to read, yielding..')
-                yield PollNeeded(sock_fd, POLLIN)
-                if self._sock is None:
-                    raise ConnectionClosed('connection has been closed locally')
-            elif not got_data:
-                log.debug('nothing more to read')
-                break
+            if got_data is None:
+                if rbuf:
+                    log.debug('nothing more to read')
+                    break
+                else:
+                    log.debug('buffer empty and nothing to read, yielding..')
+                    yield PollNeeded(self._sock.fileno(), POLLIN)
+            elif got_data == 0:
+                if self._in_remaining is READ_UNTIL_EOF:
+                    log.debug('connection closed, %d bytes in buffer', len(rbuf))
+                    self._in_remaining = len(rbuf)
+                    break
+                else:
+                    raise ConnectionClosed('server closed connection')
 
         len_ = min(len_, len(rbuf))
-        self._in_remaining -= len_
+        if self._in_remaining is not READ_UNTIL_EOF:
+            self._in_remaining -= len_
 
         if len_ < len(rbuf):
             buf = rbuf.d[rbuf.b:rbuf.b+len_]
@@ -1005,6 +1113,16 @@ class HTTPConnection:
         else:
             buf = rbuf.exhaust()
 
+        # When reading until EOF, it is possible that we read only the EOF. In
+        # this case we won't get called again, so we need to clean-up the
+        # request. This can not happen when we know the number of remaining
+        # bytes, because in this case either the check at the start of the
+        # function hits, or we can't read the remaining data and raise.
+        if len(buf) == 0:
+            assert self._in_remaining == 0
+            self._in_remaining = None
+            self._pending_requests.popleft()
+
         log.debug('done (%d bytes)', len(buf))
         return buf
 
@@ -1020,12 +1138,14 @@ class HTTPConnection:
             self._pending_requests.popleft()
             return 0
 
-        sock_fd = self._sock.fileno()
         rbuf = self._rbuf
         if not isinstance(buf, memoryview):
             buf = memoryview(buf)
-        len_ = min(len(buf), self._in_remaining)
-        log.debug('updated len_=%d', len_)
+        if self._in_remaining is READ_UNTIL_EOF:
+            len_ = len(buf)
+        else:
+            len_ = min(len(buf), self._in_remaining)
+        log.debug('set len_=%d', len_)
 
         # First use read buffer contents
         pos = min(len(rbuf), len_)
@@ -1033,11 +1153,12 @@ class HTTPConnection:
             log.debug('using buffered data')
             buf[:pos] = rbuf.d[rbuf.b:rbuf.b+pos]
             rbuf.b += pos
-            self._in_remaining -= pos
+            if self._in_remaining is not READ_UNTIL_EOF:
+                self._in_remaining -= pos
 
             # If we've read enough, return immediately
             if pos == len_:
-                log.debug('done (got all we need, %d bytes)', pos)
+                log.debug('done (buffer filled completely)')
                 return pos
 
             # Otherwise, prepare to read more from socket
@@ -1046,27 +1167,35 @@ class HTTPConnection:
 
         while True:
             log.debug('trying to read from socket')
+            if self._sock is None:
+                raise ConnectionClosed('connection has been closed locally')
             try:
                 read = self._sock.recv_into(buf[pos:len_])
+            except (ConnectionResetError, BrokenPipeError):
+                raise ConnectionClosed('connection was interrupted')
             except (socket.timeout, ssl.SSLWantReadError, BlockingIOError):
                 if pos:
-                    log.debug('done (nothing more to read, got %d bytes)', pos)
+                    log.debug('done, no additional data available')
                     return pos
                 else:
                     log.debug('no data yet and nothing to read, yielding..')
-                    yield PollNeeded(sock_fd, POLLIN)
-                    if self._sock is None:
-                        raise ConnectionClosed('connection has been closed locally')
+                    yield PollNeeded(self._sock.fileno(), POLLIN)
                     continue
 
             if not read:
-                raise ConnectionClosed('connection closed unexpectedly')
+                if self._in_remaining is READ_UNTIL_EOF:
+                    log.debug('reached EOF')
+                    self._in_remaining = 0
+                    return pos
+                else:
+                    raise ConnectionClosed('server closed connection')
+            log.debug('got %d bytes from socket', read)
 
-            log.debug('got %d bytes', read)
-            self._in_remaining -= read
+            if self._in_remaining is not READ_UNTIL_EOF:
+                self._in_remaining -= read
             pos += read
             if pos == len_:
-                log.debug('done (got all we need, %d bytes)', pos)
+                log.debug('done (buffer filled completely)')
                 return pos
 
     def _co_read_chunked(self, len_=None, buf=None):
@@ -1076,14 +1205,10 @@ class HTTPConnection:
         a `bytes-like object`. If *buf* is not `None`, reads data into *buf*.
         '''
 
-        # TODO: In readinto mode, we always need an extra sock.recv()
-        # to get the chunk trailer.. is there some way to avoid that? And
-        # maybe also put the beginning of the next chunk into the read buffer right away?
-
         log.debug('start (%s mode)', 'readinto' if buf else 'read')
         assert (len_ is None) != (buf is None)
         assert bool(len_) or bool(buf)
-        assert self._in_remaining is not None
+        assert isinstance(self._in_remaining, int)
 
         if self._in_remaining == 0:
             log.debug('starting next chunk')
@@ -1132,7 +1257,6 @@ class HTTPConnection:
 
         log.debug('reading until %s', substr)
 
-        sock_fd = self._sock.fileno()
         rbuf = self._rbuf
         sub_len = len(substr)
 
@@ -1168,11 +1292,15 @@ class HTTPConnection:
                 maxsize -= len(buf)
 
             # Refill buffer
-            while not self._try_fill_buffer():
-                log.debug('need more data, yielding')
-                yield PollNeeded(sock_fd, POLLIN)
-                if self._sock is None:
-                    raise ConnectionClosed('connection has been closed locally')
+            while True:
+                res = self._try_fill_buffer()
+                if res is None:
+                    log.debug('need more data, yielding')
+                    yield PollNeeded(self._sock.fileno(), POLLIN)
+                elif res == 0:
+                    raise ConnectionClosed('server closed connection')
+                else:
+                    break
 
         log.debug('found substr at %d', idx)
         idx += len(substr)
@@ -1191,8 +1319,11 @@ class HTTPConnection:
     def _try_fill_buffer(self):
         '''Try to fill up read buffer
 
-        Returns the number of bytes read into buffer, or `None` if no
-        data was available on the socket. May raise `ConnectionClosed`.
+        Returns the number of bytes read into buffer, or `None` if no data was
+        available on the socket. Return zero if the TCP connection has been
+        properly closed but the socket object still exists. On other problems
+        (e.g. if the socket object has been destroyed or the connection
+        interrupted), raise `ConnectionClosed`.
         '''
 
         log.debug('start')
@@ -1203,9 +1334,11 @@ class HTTPConnection:
             rbuf.b = 0
             rbuf.e = 0
 
-        # If no capacity, return
-        if rbuf.e == len(rbuf.d):
-            return 0
+        # There should be free capacity
+        assert rbuf.e < len(rbuf.d)
+
+        if self._sock is None:
+            raise ConnectionClosed('connection has been closed locally')
 
         try:
             len_ = self._sock.recv_into(memoryview(rbuf.d)[rbuf.e:])
@@ -1213,10 +1346,7 @@ class HTTPConnection:
             log.debug('done (nothing ready)')
             return None
         except (ConnectionResetError, BrokenPipeError):
-            len_ = 0
-
-        if not len_:
-            raise ConnectionClosed('connection closed unexpectedly')
+            raise ConnectionClosed('connection was interrupted')
 
         rbuf.e += len_
         log.debug('done (got %d bytes)', len_)
@@ -1228,14 +1358,14 @@ class HTTPConnection:
         rbuf = self._rbuf
         if len_ > len(rbuf.d):
             raise ValueError('Requested more bytes than buffer has capacity')
-        sock_fd = self._sock.fileno()
         while len(rbuf) < len_:
             if len(rbuf.d) - rbuf.b < len_:
                 self._rbuf.compact()
-            if not self._try_fill_buffer():
-                yield PollNeeded(sock_fd, POLLIN)
-                if self._sock is None:
-                    raise ConnectionClosed('connection has been closed locally')
+            res = self._try_fill_buffer()
+            if res is None:
+                yield PollNeeded(self._sock.fileno(), POLLIN)
+            elif res == 0:
+                raise ConnectionClosed('server closed connection')
 
     def readall(self):
         '''placeholder, will be replaced dynamically'''
@@ -1244,9 +1374,6 @@ class HTTPConnection:
     def co_readall(self):
         '''Read and return complete response body'''
 
-        if self._sock is None:
-            raise ConnectionClosed('connection has been closed locally')
-
         if self._in_remaining is None:
             return b''
 
@@ -1281,6 +1408,14 @@ class HTTPConnection:
             log.debug('discarding %d bytes', len_)
         log.debug('done')
 
+    def reset(self):
+        '''Reset HTTP connection
+
+        This method resets the status of the HTTP connection after an exception
+        has occured. Any cached data and pending responses are discarded.
+        '''
+        self.disconnect()
+
     def disconnect(self):
         '''Close HTTP connection'''
 
@@ -1363,36 +1498,82 @@ def eval_coroutine(crt, timeout=None):
     except StopIteration as exc:
         return exc.value
 
+def create_socket(address):
+    '''Create socket connected to *address**
+
+    Use `socket.create_connection` to create a connected socket and return
+    it. If a DNS related exception is raised, capture it and attempt to
+    determine if the dns server is not reachable, or if the host name could
+    not be resolved. Then raise either `NoSuchAddress` or
+    `NameResolutionError` as appropriate.
+
+    To distinguish between an unresolvable hostname and a problem with the
+    DNS server, attempt to resolve the addresses in `DNS_TEST_HOSTNAMES`. If
+    at least one test hostname can be resolved, assume that the DNS server
+    is available and that *address* can not be resolved.
+    '''
+
+    try:
+        return socket.create_connection(address)
+
+    # The exception unfortunately does not help us to distinguish between
+    # permanent and temporary problems. See:
+    # https://stackoverflow.com/questions/24855168/
+    # https://stackoverflow.com/questions/24855669/
+    except (socket.gaierror, socket.herror) as exc:
+        if exc.errno not in (socket.EAI_AGAIN, socket.EAI_NONAME):
+            raise
+
+    # Try to resolve test hosts
+    for (hostname, port) in DNS_TEST_HOSTNAMES:
+        try:
+            socket.getaddrinfo(hostname, port)
+        except (socket.gaierror, socket.herror) as exc:
+            if exc.errno not in (socket.EAI_AGAIN, socket.EAI_NONAME):
+                raise
+            # Not reachable, try next one
+        else:
+            # Reachable, now try to resolve original address
+            # again (maybe dns was only down briefly)
+            break
+    else:
+        # No host was reachable
+        raise DNSUnavailable(address[0])
+
+    # Try to connect to original host again
+    try:
+        return socket.create_connection(address)
+    except (socket.gaierror, socket.herror) as exc:
+        if exc.errno not in (socket.EAI_AGAIN, socket.EAI_NONAME):
+            raise
+        raise HostnameNotResolvable(address[0])
+
+
 def is_temp_network_error(exc):
     '''Return true if *exc* represents a potentially temporary network problem
 
-    For problems with name resolution, the exception generally does not contain
-    enough information to distinguish between an unresolvable hostname and a
-    problem with the DNS server. In this case, this function attempts to resolve
-    the addresses in `DNS_TEST_HOSTNAMES`. A name resolution problem is
-    considered permanent if at least one test hostname can be resolved, and
-    temporary if none of the test hostnames can be resolved.
+    DNS resolution errors (`socket.gaierror` or `socket.herror`) are considered
+    permanent, because `HTTPConnection` employs a heuristic to convert these
+    exceptions to `HostnameNotResolvable` or `DNSUnavailable` instead.
     '''
 
     if isinstance(exc, (socket.timeout, ConnectionError, TimeoutError, InterruptedError,
                         ConnectionClosed, ssl.SSLZeroReturnError, ssl.SSLEOFError,
-                        ssl.SSLSyscallError, ConnectionTimedOut)):
+                        ssl.SSLSyscallError, ConnectionTimedOut, DNSUnavailable)):
         return True
 
-    # The exception also unfortunately does not help us to distinguish between
-    # permanent and temporary problems. See:
-    # https://stackoverflow.com/questions/24855168/
-    # https://stackoverflow.com/questions/24855669/
-    elif (isinstance(exc, (socket.gaierror, socket.herror))
-          and exc.errno in (socket.EAI_AGAIN, socket.EAI_NONAME)):
-        for (hostname, port) in DNS_TEST_HOSTNAMES:
+    elif isinstance(exc, OSError):
+        # We have to be careful when retrieving errno codes, because
+        # not all of them may exist on every platform.
+        for errcode in ('EHOSTDOWN', 'EHOSTUNREACH', 'ENETDOWN',
+                        'ENETRESET', 'ENETUNREACH', 'ENOLINK',
+                        'ENONET', 'ENOTCONN', 'ENXIO', 'EPIPE',
+                        'EREMCHG', 'ESHUTDOWN', 'ETIMEDOUT'):
             try:
-                socket.getaddrinfo(hostname, port)
-            except (socket.gaierror, socket.herror):
+                if getattr(errno, errcode) == exc.errno:
+                    return True
+            except AttributeError:
                 pass
-            else:
-                return False
-        return True
 
     return False
 
